/*!
  \file
  \brief oHǏ]

  \author Satofumi KAMIMURA

  $Id: follow_control.c 1219 2009-08-14 02:59:46Z satofumi $
*/

#include "follow_control.h"
#include "position_calculate.h"
#include "isqrt.h"
#include "iatan2.h"

#include <stdio.h>
#include <stdlib.h>

enum {
  X = 0,
  Y = 1,
};


void follow_initialize(follow_t *follow, long acceleration)
{
  follow->reset = 0;
  follow->target_velocity = 0;
  follow->control_velocity = 0;
  follow->control_acceleration = acceleration;
}


void follow_reset(follow_t *follow)
{
  (void)follow;

  // !!!
}


long follow_velocityControl(follow_t *follow)
{
  long target_velocity = follow->target_velocity;
  long control_velocity = follow->control_velocity;
  long acceleration = follow->control_acceleration;

  if ((control_velocity + (acceleration >> FOLLOW_SHIFT)) < target_velocity) {
    control_velocity += acceleration >> FOLLOW_SHIFT;

  } else if ((control_velocity
              - (acceleration >> FOLLOW_SHIFT)) > target_velocity) {
    control_velocity -= acceleration >> FOLLOW_SHIFT;

  } else {
    control_velocity = target_velocity;
  }

  follow->control_velocity = control_velocity;
  return control_velocity >> FOLLOW_SHIFT;
}


long follow_positionControl(follow_t *follow, long difference)
{
  // v^2 = 2al vZxpđxs
  int sign = (difference >= 0) ? +1 : -1;
  if (sign < 0) {
    difference = -difference;
  }

  // P isqrt() ŌvZƕϐ bit  32 bit ̏ꍇɂȂ
  long velocity =
    isqrt(follow->control_acceleration >> (FOLLOW_SHIFT - 1)) *
    isqrt(difference);
  velocity <<= FOLLOW_SHIFT;

  if (sign < 0) {
    velocity = -velocity;
  }
  follow->target_velocity = velocity;

  return follow_velocityControl(follow);
}


static long follow_direction(long abs_mm, long follow_radius, int sign)
{
  enum { LengthOffset = 16 };

  long dir16;
  long y;

  // UȂ߂̕␳
  if (abs_mm > LengthOffset) {
    abs_mm -= LengthOffset;
  } else {
    abs_mm = 0;
  }

  // ~ʂɐڂ钼̌XɁA𐧌䂷
  // theta = atan2(y, sqrt(y * (2 * r - y)));
  y = isqrt(abs_mm * ((follow_radius << 1) - abs_mm));
  dir16 = iatan2(y, follow_radius - abs_mm);
  if (sign > 0) {
    dir16 = -dir16;
  }

  return dir16;
}


long follow_line(follow_t *rotate, const position_t *position,
                 const path_t *path)
{
  // position->mm[Y] Ƃ̋ɉČ̐s
  // w蒼ɑ΂āA~ʂ̋OՂ`ċߕt悤ɂ

  long follow_radius = path->follow_radius;
  long mm = position->mm[Y];
  int sign = (mm >= 0) ? +1 : -1;
  long abs_mm = (sign > 0) ? mm : -mm;
  long dir16;
  long diff;

  // oHǏ]̋ȗĂA̕Ɍ悤ɐ䂷
  if (abs_mm > follow_radius) {
    dir16 = -sign * (65536 >> 2);

  } else {
    dir16 = follow_direction(abs_mm, follow_radius, sign);
  }

  diff = position_dir16diff(dir16, position->dir16);
  return follow_positionControl(rotate, diff);
}


long follow_circle(follow_t *rotate, const position_t *position,
                   const path_t *path)
{
  long follow_radius = path->follow_radius;
  long x = position->mm[X];
  long y = position->mm[Y];

  // Ǐ]~ʂƂ̋vZ
  long r = isqrt((x * x) + (y * y));
  int circle_sign = (path->circle_radius > 0) ? +1 : -1;
  long mm = path->circle_radius - (circle_sign * r);
  long abs_mm = (mm > 0) ? mm : -mm;
  int mm_sign = (mm > 0) ? +1 : -1;
  long dir16;
  long diff;

  //fprintf(stderr, "%ld, ", mm);

  // oHǏ]̋ȗĂA~ʂ̕Ɍ悤ɐ䂷
  if (abs_mm > follow_radius) {
    dir16 = iatan2(y, x) + ((mm_sign * circle_sign < 0) ? (65536 >> 1) : 0);
    //fprintf(stderr, "<%ld>, ", (360 * dir16) >> 16);
  } else {
    //fprintf(stderr, "<%d>, ", mm_sign);
    dir16 = follow_direction(abs_mm, follow_radius, mm_sign);
    //fprintf(stderr, "<%ld>, ", (360 * dir16) >> 16);

    // ~ʕ̌Xɐ䂷
    dir16 += iatan2(y, x) + (circle_sign * (65536 >> 2));
  }

  // ]pxɑ␳s
  // !!! ix֌WĂ͂
  // !!!

  diff = position_dir16diff(dir16, position->dir16);
  //fprintf(stderr, "%ld, %d, %ld\n", dir16, position->dir16, diff);
  return follow_positionControl(rotate, diff);
}


void spin(void)
{
  // !!!
}
